let print_boolean a = match a with
	| true 	-> print_string("true")
	| false -> print_string("false")
;;


(* Exercice 1 *)
type style = 
  | Gras
  | Italique
  | Souligne
     
;;

(* Exercice 2 *)
type text = 
  | Texte of string 
  | Style of style * text
  | Concat of text list
;;

let print_style s = match s with
  | Gras -> print_string("Bold")
  | Italique -> print_string("Italics")
  | Souligne -> print_string("Souligne")
;;

let rec print_text t = match t with
  | Texte(a) ->
    print_string(a ^ " ")  
  | Style(style,text) ->
    print_text(text)
  | Concat(l) -> 
    List.iter print_text l 
;;

let rec print_list l = match l with
  | [] -> print_string("")
  | a::b -> print_text(a) ; print_list b
;;

let rec print_bloc_test b = match b with
  | Texte(a) -> 
    print_string("Text " ^ a ^ " ")
  | Style(style, text) -> 
    print_string("Style (");
    print_style(style);
    print_string(", ");
    print_bloc_test(text);
    print_string(")")
  | Concat(list) -> 
    print_string("Concat [");
    List.iter print_bloc_test list;
    print_string( "] ")
;;
   

(* Exercice 3*)
let rec paren p = match p with
  | Texte(a) -> 
    Texte(a)
  | Concat(list) ->
    Concat(List.map paren list)
  | Style(style, text) ->
    match style with
      | Italique ->
	Style(style, Concat( [Texte("(")] @ [paren(text)] @ [Texte(")")] ) )
      | _ -> 
	paren(text)
;;
 

(* Exercice 4 *)
let rec plain p = match p with
  | Texte(a) -> 
    Texte(a)
  | Concat(list) ->
    Concat(List.map plain list)
  | Style(s, t) ->
    plain(t)
;;


(* Exercice 5 *)
let rec nobold n = match n with
  | Texte(a) ->
    Texte(a)
  | Concat(list) ->
    Concat(List.map nobold list )
  | Style(s, t) ->
    match s with
      | Italique 
      | Souligne ->
	Style( s, nobold t)
      | _ -> 
	Style(s, Texte(""))
;;


(* Exercice 6 *)
let rec nobold3 n = match n with
  | Texte(a) ->
    Texte(a)
  | Concat(list) ->
    Concat(List.map nobold3 list)
  | Style(s, t) ->
    match s with
      | Italique ->
	Style(s, nobold3(t))
      | _ ->
	Texte("")
;;
let rec nobold2 n = match n with
  | Texte(a) ->
    Texte(a)
  | Concat(list) ->
      Concat(List.map nobold2 list)
  | Style(s, t)-> 
    match s with
      | Italique | Souligne ->
	Style(s, nobold2(t) )
      | Gras ->
	nobold3(t)
;; 

(* Exercice 7 *)
let rec italics i= match i with
  | Texte(a) ->
    []
  | Concat(list) ->
     List.fold_left (fun x y -> x @ italics y) [] list
  | Style(s, t)->
    match s with
      | Italique ->
	[t]
      | _ ->
	italics t
;;

(*Exercice 8 *)
let rec to_html t = match t with
  | Texte(a) ->
    a
  | Concat(list) ->
    List.fold_left (fun x y -> x ^ to_html y) "" list
  | Style(s, t) ->
    match s with
      | Italique ->
	"<i>" ^ to_html (t) ^ "</i>"
      | Gras ->
	"<b>" ^ to_html (t) ^ "</b>"
      | Souligne ->
	"<u>" ^ to_html (t) ^ "</u>"
;;

(* Exercice 9 *)
(* 1 *)
let rec map f t = match t with
      | Texte(a) ->
	Texte(a)
      | Concat(list) ->
	Concat(f list)
      | Style(s, b)->
	Style(s, map f b)
;;

(* 2 *)
let rec remove texte = match texte with
  | (Texte(t) as tex)-> tex
  | Concat(list) -> 
    Concat(List.fold_left 
	     ( fun x y-> match y with 
	       | Texte("") 
	       | Concat([]) -> x 
	       | _-> x @ [remove y]) 
	     [] list)
  | Style(s, t)-> Style(s, remove t)
;;

(* 3 *)
let rec flatten texte = match texte with
  | Texte(t) -> Texte(t)
  | Concat(list) ->
    Concat( List.fold_left 
	      ( fun x y -> match y with
		| Concat(list2) -> x @ (List.map flatten list2)
		| _ ->  x @ [y]
	      )
	      [] list)
  | Style(s, t) ->
    Style(s, flatten t)
;;

(* 4 *)
let rec adjacent list = match list with
  | [] -> [] 
  | Texte(a)::Texte(b)::c ->
    [Texte(a ^ " " ^ b)]@ adjacent c
  | a::b -> [concat a] @ adjacent b
and concat texte = match texte with
  | Texte(t) -> Texte(t)
  | Style(s, t) -> Style(s, concat t)
  | Concat(liste) -> Concat(adjacent liste)
;;

(* 5 *)
let normalize n = 
  concat (concat (flatten (remove n)))
;;

(* Exercice 11 *)
type document =
  | Paragraphe of text
  | Section of text * document list

(* Exercice 12 *)
let rec map_text f d = match d with
  | Paragraphe(p) ->
    Paragraphe(f p)
  | Section(t, dl) ->
    Section(f t,List.map (fun x -> map_text f x) (dl) ) 
;;

(* Exercice 13 *)
let rec to_html3 html h= match html with
  | Paragraphe(p) -> 
    to_html p
  | Section(t, dl) -> 
    "<h"^ (string_of_int h)^ ">" ^ (to_html t) ^ "</h" ^(string_of_int h )^ ">" ^ (List.fold_left (fun x y-> x ^ to_html3 y (succ h)) "" dl ) 
;;

(* Exercice 14 *)
type toc = Toc of text * toc list;;
let rec toc_of_document tod = match tod with
  | Paragraphe(p) -> Toc(p, [])
  | Section(t, dl) -> Toc(t, List.map (toc_of_document) dl)
;;

let rec toc_to_html tth = match tth with
  | Paragraphe(p) -> 
    "<li>" ^ to_html p ^ "</li>"
  | Section(t, dl) ->
    (to_html t) ^"<ul>" ^ ( List.fold_left (fun x y-> x ^ toc_to_html y) "" dl ) ^ "</ul>"

(* Exercice 15 *)
let rec numerote_aux n_a b n = match n_a with
  | Paragraphe(p) -> Paragraphe(p)
  | Section(t, dl) -> 
    Section(Concat[t;Texte(string_of_int n);Texte(".")],List.map (fun x -> numerote_aux x (b ^ "." ^ (string_of_int n)) (succ n)) dl) 
;; 
let rec numerote n = numerote_aux n "" 1;;

let rec to_html h = to_html3 h 1;;

let bloc = Concat[Style(Italique, Texte("ee"));Texte("test1");Texte("test2");Texte("test3");Texte("a");Concat[Texte("test4");Texte("test5")]];;

(*
let bloc = Concat[Texte("test1");Texte("test2");Concat[Texte("test3")]];;
*)

print_bloc_test (normalize bloc);;
