type unop = Cos | Sin;;
type binop = Plus | Moins | Fois;;
type expr = | Num of float
	    | Var of string
	    | Unop of unop * expr
	    | Binop of binop * expr * expr
;;

(* Exercice 1 *)
let string_of_binop b = match b with
  | Plus  -> " + "
  | Moins -> " - "
  | Fois  -> " * "
;;

let string_of_unop u = match u with
  | Cos -> "cos"
  | Sin -> "sin"
;;

(* Exercice 2 *)
let rec string_of_expr e = match e with
  | Num(f) 
    -> string_of_float(f)
  | Var(v) 
    -> v
  | Unop(u, e1) 
    -> string_of_unop(u) ^ "( " ^ string_of_expr(e1) ^ " )"
  | Binop(b, e1, e2)
    -> "( " ^ string_of_expr(e1) ^ string_of_binop(b) ^ string_of_expr(e2) ^ " )"
;;

(* Exercice 3 *)
let val_unop u = match u with
  | Cos -> (fun x -> cos(x))
  | Sin -> (fun x -> cos(x))
;;

let val_binop b = match b with
  | Plus  -> (fun x y -> x +. y )
  | Moins -> (fun x y -> x -. y)
  | Fois  -> (fun x y -> x *. y)
;;

(* Exercice 4 *)
let rec eval_expr lalist expression = match expression with
  | Num(f)           -> f
  | Var(v)           -> List.assoc v lalist
  | Unop(u, e1)      -> val_unop u ( eval_expr lalist e1)
  | Binop(u, e1, e2) -> val_binop u (eval_expr lalist e1) (eval_expr lalist e2)
;;

(* Exercice 5 *) 


(* Exercice 6 *)
let rec eval_neutres e = match e with 
  | Num(f)           -> Num(f)
  | Var(v)           -> Var(v)
  | Unop(u, e1)      -> Unop(u, e1)
  | Binop(u, e1, e2) -> 
    match (u,e1,e2) with 
      | (Fois, Num(0.), x) -> Num(0.) 
      | (Fois, y, Num(0.)) -> Num(0.)
      | (Fois, Num(1.), x) -> x
      | (Fois, y, Num(1.)) -> y
      | (Moins,y, Num(0.)) -> y
      | (Moins, x, y) -> if(x == y ) then Num(0.) else Binop(Moins,x,y)
      | (Plus, x, Num(0.)) -> x
      | (Plus, Num(0.),x) -> x
      | _ -> Binop(u,e1,e2)

;;

(* Exercice 7 *)
let point_fixe f a = f(f(a));;

let simplifier e = 
  point_fixe (eval_neutres) e
;;

(* Exercice 8 *)
let rec deriv_expr s e= match e with
      | Num(f) -> Num(0.)
      | Var(v) -> if compare s v == 0 then Num(1.) else Num(0.)
      | Unop(u, e1) -> 
	if( u == Cos ) then 
	  Binop(Moins, Num(0.), Binop(Fois, deriv_expr s e1, Unop(Sin,e1))) 
	else
	  Binop(Fois, deriv_expr s e1, Unop(Cos, e1))
      | Binop(u, e1, e2) -> 
	match u with 
	  | Plus -> Binop(u, deriv_expr s e1, deriv_expr s e2)
	  | Fois -> Binop(Plus, Binop(Fois, deriv_expr s e1, e2), (Binop(Fois, e1, deriv_expr s e2)))
	  | _ -> Binop(u, e1, e2)
;;
let e0 = Binop (Fois,
		Unop (Cos, Binop(Plus, Var "x", Num 0.5)),
		Binop(Moins, Var "theta", Num 2.));;

print_float(eval_expr [("x",-0.5);("theta",4.)] e0 );;
